#
# Copyright (c) 2021-2022 Carsten Igel.
#
# This file is part of pdm-bump 
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license. 
# Refer to LICENSE for more information
#

import unittest
from typing import List, Tuple, Callable

from pdm_bump.action import (
    PreviewMismatchError,
    VersionModifier,
    MajorIncrementingVersionModifier,
    MinorIncrementingVersionModifier,
    MicroIncrementingVersionModifier,
    AlphaIncrementingVersionModifier,
    BetaIncrementingVersionModifier,
    ReleaseCandidateIncrementingVersionModifier,
    FinalizingVersionModifier,
    EpochIncrementingVersionModifier,
    DevelopmentVersionIncrementingVersionModifier,
    PostVersionIncrementingVersionModifier,
)
from pdm_bump.version import Version

class ActionTest(unittest.TestCase):
    CREATE_NEXT_VERSION_PARAMS: List[Tuple[str, str, str, Callable[[Version], Version]]] = [
        ("Increment major parts with removing pre-parts and no pre-parts", "1.0.0", "2.0.0", lambda v: MajorIncrementingVersionModifier(v, True)),
        ("Increment major parts with removing pre-parts and alpha pre-parts", "1.0.0a1", "2.0.0", lambda v: MajorIncrementingVersionModifier(v, True)),
        ("Increment major parts with removing pre-parts and beta pre-parts", "1.0.0b1", "2.0.0", lambda v: MajorIncrementingVersionModifier(v, True)),
        ("Increment major parts with removing pre-parts and rc pre-parts", "1.0.0rc1", "2.0.0", lambda v: MajorIncrementingVersionModifier(v, True)),
        ("Increment major parts with removing pre-parts and dev parts", "1.0.0dev1", "2.0.0", lambda v: MajorIncrementingVersionModifier(v, True)),
        ("Increment major parts with removing pre-parts and post parts", "1.0.0post1", "2.0.0", lambda v: MajorIncrementingVersionModifier(v, True)),
        ("Increment major parts with removing pre-parts and local parts", "1.0.0+local", "2.0.0", lambda v: MajorIncrementingVersionModifier(v, True)),
        ("Increment major parts without removing pre-parts and no pre-parts", "1.0.0", "2.0.0", lambda v: MajorIncrementingVersionModifier(v, False)),
        ("Increment major parts without removing pre-parts and alpha pre-parts", "1.0.0a1", "2.0.0a1", lambda v: MajorIncrementingVersionModifier(v, False)),
        ("Increment major parts without removing pre-parts and beta pre-parts", "1.0.0b1", "2.0.0b1", lambda v: MajorIncrementingVersionModifier(v, False)),
        ("Increment major parts without removing pre-parts and rc pre-parts", "1.0.0rc1", "2.0.0rc1", lambda v: MajorIncrementingVersionModifier(v, False)),
        ("Increment major parts without removing pre-parts and dev parts", "1.0.0dev1", "2.0.0dev1", lambda v: MajorIncrementingVersionModifier(v, False)),
        ("Increment major parts without removing pre-parts and post parts", "1.0.0post1", "2.0.0post1", lambda v: MajorIncrementingVersionModifier(v, False)),
        ("Increment major parts without removing pre-parts and local parts", "1.0.0+local", "2.0.0+local", lambda v: MajorIncrementingVersionModifier(v, False)),

        ("Increment minor parts with removing pre-parts and no pre-parts", "1.0.0", "1.1.0", lambda v: MinorIncrementingVersionModifier(v, True)),
        ("Increment minor parts with removing pre-parts and alpha pre-parts", "1.0.0a1", "1.1.0", lambda v: MinorIncrementingVersionModifier(v, True)),
        ("Increment minor parts with removing pre-parts and beta pre-parts", "1.0.0b1", "1.1.0", lambda v: MinorIncrementingVersionModifier(v, True)),
        ("Increment minor parts with removing pre-parts and rc pre-parts", "1.0.0rc1", "1.1.0", lambda v: MinorIncrementingVersionModifier(v, True)),
        ("Increment minor parts with removing pre-parts and dev parts", "1.0.0dev1", "1.1.0", lambda v: MinorIncrementingVersionModifier(v, True)),
        ("Increment minor parts with removing pre-parts and post parts", "1.0.0post1", "1.1.0", lambda v: MinorIncrementingVersionModifier(v, True)),
        ("Increment minor parts with removing pre-parts and local parts", "1.0.0+local", "1.1.0", lambda v: MinorIncrementingVersionModifier(v, True)),
        ("Increment minor parts without removing pre-parts and no pre-parts", "1.0.0", "1.1.0", lambda v: MinorIncrementingVersionModifier(v, False)),
        ("Increment minor parts without removing pre-parts and alpha pre-parts", "1.0.0a1", "1.1.0a1", lambda v: MinorIncrementingVersionModifier(v, False)),
        ("Increment minor parts without removing pre-parts and beta pre-parts", "1.0.0b1", "1.1.0b1", lambda v: MinorIncrementingVersionModifier(v, False)),
        ("Increment minor parts without removing pre-parts and rc pre-parts", "1.0.0rc1", "1.1.0rc1", lambda v: MinorIncrementingVersionModifier(v, False)),
        ("Increment minor parts without removing pre-parts and dev parts", "1.0.0dev1", "1.1.0dev1", lambda v: MinorIncrementingVersionModifier(v, False)),
        ("Increment minor parts without removing pre-parts and post parts", "1.0.0post1", "1.1.0post1", lambda v: MinorIncrementingVersionModifier(v, False)),
        ("Increment minor parts without removing pre-parts and local parts", "1.0.0+local", "1.1.0+local", lambda v: MinorIncrementingVersionModifier(v, False)),

        ("Increment micro parts with removing pre-parts and no pre-parts", "1.0.0", "1.0.1", lambda v: MicroIncrementingVersionModifier(v, True)),
        ("Increment micro parts with removing pre-parts and alpha pre-parts", "1.0.0a1", "1.0.1", lambda v: MicroIncrementingVersionModifier(v, True)),
        ("Increment micro parts with removing pre-parts and beta pre-parts", "1.0.0b1", "1.0.1", lambda v: MicroIncrementingVersionModifier(v, True)),
        ("Increment micro parts with removing pre-parts and rc pre-parts", "1.0.0rc1", "1.0.1", lambda v: MicroIncrementingVersionModifier(v, True)),
        ("Increment micro parts with removing pre-parts and dev parts", "1.0.0dev1", "1.0.1", lambda v: MicroIncrementingVersionModifier(v, True)),
        ("Increment micro parts with removing pre-parts and post parts", "1.0.0post1", "1.0.1", lambda v: MicroIncrementingVersionModifier(v, True)),
        ("Increment micro parts with removing pre-parts and local parts", "1.0.0+local", "1.0.1", lambda v: MicroIncrementingVersionModifier(v, True)),
        ("Increment micro parts without removing pre-parts and no pre-parts", "1.0.0", "1.0.1", lambda v: MicroIncrementingVersionModifier(v, False)),
        ("Increment micro parts without removing pre-parts and alpha pre-parts", "1.0.0a1", "1.0.1a1", lambda v: MicroIncrementingVersionModifier(v, False)),
        ("Increment micro parts without removing pre-parts and beta pre-parts", "1.0.0b1", "1.0.1b1", lambda v: MicroIncrementingVersionModifier(v, False)),
        ("Increment micro parts without removing pre-parts and rc pre-parts", "1.0.0rc1", "1.0.1rc1", lambda v: MicroIncrementingVersionModifier(v, False)),
        ("Increment micro parts without removing pre-parts and dev parts", "1.0.0dev1", "1.0.1dev1", lambda v: MicroIncrementingVersionModifier(v, False)),
        ("Increment micro parts without removing pre-parts and post parts", "1.0.0post1", "1.0.1post1", lambda v: MicroIncrementingVersionModifier(v, False)),
        ("Increment micro parts without removing pre-parts and local parts", "1.0.0+local", "1.0.1+local", lambda v: MicroIncrementingVersionModifier(v, False)),

        ("Increment alpha parts without incrementing micro and providing no pre-release part", "1.0.0", "1.0.0a1", lambda v: AlphaIncrementingVersionModifier(v, False)),
        ("Increment alpha parts without incrementing micro and providing a pre-release part", "1.0.0a1", "1.0.0a2", lambda v: AlphaIncrementingVersionModifier(v, False)),
        ("Increment alpha parts with incrementing micro and providing no pre-release part", "1.0.0", "1.0.1a1", lambda v: AlphaIncrementingVersionModifier(v, True)),
        ("Increment alpha parts with incrementing micro and providing a pre-release part", "1.0.0a1", "1.0.0a2", lambda v: AlphaIncrementingVersionModifier(v, True)),

        ("Increment beta parts without incrementing micro and providing no pre-release part", "1.0.0", "1.0.0b1", lambda v: BetaIncrementingVersionModifier(v, False)),
        ("Increment beta parts without incrementing micro and providing a pre-release part", "1.0.0b1", "1.0.0b2", lambda v: BetaIncrementingVersionModifier(v, False)),
        ("Increment beta parts without incrementing micro and providing a pre-release part", "1.0.0a2", "1.0.0b1", lambda v: BetaIncrementingVersionModifier(v, False)),
        ("Increment beta parts with incrementing micro and providing no pre-release part", "1.0.0", "1.0.1b1", lambda v: BetaIncrementingVersionModifier(v, True)),
        ("Increment beta parts with incrementing micro and providing a pre-release part", "1.0.0b1", "1.0.0b2", lambda v: BetaIncrementingVersionModifier(v, True)),
        ("Increment beta parts with incrementing micro and providing no pre-release part", "1.0.0a2", "1.0.0b1", lambda v: BetaIncrementingVersionModifier(v, True)),

        ("Increment rc parts without incrementing micro and providing no pre-release part", "1.0.0", "1.0.0rc1", lambda v: ReleaseCandidateIncrementingVersionModifier(v, False)),
        ("Increment rc parts without incrementing micro and providing a pre-release part", "1.0.0rc1", "1.0.0rc2", lambda v: ReleaseCandidateIncrementingVersionModifier(v, False)),
        ("Increment rc parts without incrementing micro and providing no pre-release part", "1.0.0a2", "1.0.0rc1", lambda v: ReleaseCandidateIncrementingVersionModifier(v, False)),
        ("Increment rc parts without incrementing micro and providing no pre-release part", "1.0.0b2", "1.0.0rc1", lambda v: ReleaseCandidateIncrementingVersionModifier(v, False)),
        ("Increment rc parts with incrementing micro and providing no pre-release part", "1.0.0", "1.0.1rc1", lambda v: ReleaseCandidateIncrementingVersionModifier(v, True)),
        ("Increment rc parts with incrementing micro and providing a pre-release part", "1.0.0rc1", "1.0.0rc2", lambda v: ReleaseCandidateIncrementingVersionModifier(v, True)),
        ("Increment rc parts with incrementing micro and providing no pre-release part", "1.0.0a2", "1.0.0rc1", lambda v: ReleaseCandidateIncrementingVersionModifier(v, True)),
        ("Increment rc parts with incrementing micro and providing a pre-release part", "1.0.0b2", "1.0.0rc1", lambda v: ReleaseCandidateIncrementingVersionModifier(v, True)),

        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3", "1!1.2.3", lambda v: EpochIncrementingVersionModifier(v, False, False)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3a4", "1!1.2.3a4", lambda v: EpochIncrementingVersionModifier(v, False, False)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3b4", "1!1.2.3b4", lambda v: EpochIncrementingVersionModifier(v, False, False)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3rc4", "1!1.2.3rc4", lambda v: EpochIncrementingVersionModifier(v, False, False)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3dev5", "1!1.2.3dev5", lambda v: EpochIncrementingVersionModifier(v, False, False)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3post6", "1!1.2.3post6", lambda v: EpochIncrementingVersionModifier(v, False, False)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3+local7", "1!1.2.3+local7", lambda v: EpochIncrementingVersionModifier(v, False, False)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3b4post3dev6+local7", "1!1.2.3b4post3dev6+local7", lambda v: EpochIncrementingVersionModifier(v, False, False)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3", "1!1.2.3", lambda v: EpochIncrementingVersionModifier(v, True, False)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3a4", "1!1.2.3", lambda v: EpochIncrementingVersionModifier(v, True, False)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3b4", "1!1.2.3", lambda v: EpochIncrementingVersionModifier(v, True, False)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3rc4", "1!1.2.3", lambda v: EpochIncrementingVersionModifier(v, True, False)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3dev5", "1!1.2.3", lambda v: EpochIncrementingVersionModifier(v, True, False)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3post6", "1!1.2.3", lambda v: EpochIncrementingVersionModifier(v, True, False)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3+local7", "1!1.2.3", lambda v: EpochIncrementingVersionModifier(v, True, False)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3b4post3dev6+local7", "1!1.2.3", lambda v: EpochIncrementingVersionModifier(v, True, False)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3", "1!1.0.0", lambda v: EpochIncrementingVersionModifier(v, False, True)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3a4", "1!1.0.0", lambda v: EpochIncrementingVersionModifier(v, False, True)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3b4", "1!1.0.0", lambda v: EpochIncrementingVersionModifier(v, False, True)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3rc4", "1!1.0.0", lambda v: EpochIncrementingVersionModifier(v, False, True)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3dev5", "1!1.0.0", lambda v: EpochIncrementingVersionModifier(v, False, True)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3post6", "1!1.0.0", lambda v: EpochIncrementingVersionModifier(v, False, True)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3+local7", "1!1.0.0", lambda v: EpochIncrementingVersionModifier(v, False, True)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3b4post3dev6+local7", "1!1.0.0", lambda v: EpochIncrementingVersionModifier(v, False, True)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3", "1!1.0.0", lambda v: EpochIncrementingVersionModifier(v, True, True)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3a4", "1!1.0.0", lambda v: EpochIncrementingVersionModifier(v, True, True)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3b4", "1!1.0.0", lambda v: EpochIncrementingVersionModifier(v, True, True)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3rc4", "1!1.0.0", lambda v: EpochIncrementingVersionModifier(v, True, True)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3dev5", "1!1.0.0", lambda v: EpochIncrementingVersionModifier(v, True, True)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3post6", "1!1.0.0", lambda v: EpochIncrementingVersionModifier(v, True, True)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3+local7", "1!1.0.0", lambda v: EpochIncrementingVersionModifier(v, True, True)),
        ("Increment epoch without an epoch and without resetting version or removing any version part", "1.2.3b4post3dev6+local7", "1!1.0.0", lambda v: EpochIncrementingVersionModifier(v, True, True)),

        ("Remove non-final parts", "1.2.3", "1.2.3", lambda v: FinalizingVersionModifier(v)),
        ("Remove non-final parts", "1.2.3a1", "1.2.3", lambda v: FinalizingVersionModifier(v)),
        ("Remove non-final parts", "1.2.3b2", "1.2.3", lambda v: FinalizingVersionModifier(v)),
        ("Remove non-final parts", "1.2.3rc3", "1.2.3", lambda v: FinalizingVersionModifier(v)),
        ("Remove non-final parts", "1.2.3-dev1", "1.2.3", lambda v: FinalizingVersionModifier(v)),
        ("Remove non-final parts", "1.2.3-post4", "1.2.3", lambda v: FinalizingVersionModifier(v)),
        ("Remove non-final parts", "1.2.3+local8", "1.2.3", lambda v: FinalizingVersionModifier(v)),
        ("Remove non-final parts", "1.2.3-b4-post6-dev8+local9", "1.2.3", lambda v: FinalizingVersionModifier(v)),
    ]  # TODO Test Dev and Post Incrementing modifier
    CREATE_NEXT_VERSION_ERROR_PARAMS: List[Tuple[str, str, Callable[[Version], VersionModifier]]] = [
        ("Increment alpha version if beta is present", "1.2.3b1", lambda v: AlphaIncrementingVersionModifier(v, False)),
        ("Increment alpha version if beta is present", "1.2.3b1", lambda v: AlphaIncrementingVersionModifier(v, True)),
        ("Increment alpha version if rc is present", "1.2.3rc1", lambda v: AlphaIncrementingVersionModifier(v, False)),
        ("Increment alpha version if rc is present", "1.2.3rc1", lambda v: AlphaIncrementingVersionModifier(v, True)),
        ("Increment beta version if rc is present", "1.2.3rc1", lambda v: BetaIncrementingVersionModifier(v, False)),
        ("Increment beta version if rc is present", "1.2.3rc1", lambda v: BetaIncrementingVersionModifier(v, True)),
    ]

    def test_create_next_version_success(self) -> None:
        for message, current_version_str, expected_version_str, factory in self.CREATE_NEXT_VERSION_PARAMS:
            with self.subTest(message, current=current_version_str, expected=expected_version_str):
                current: Version = Version.from_string(current_version_str)
                expected: Version = Version.from_string(expected_version_str)

                command: VersionModifier = factory(current)

                modified: Version = command.create_new_version()
                self.assertEqual(modified, expected)

    def test_create_next_version_fail(self) -> None:
        for message, current_version_str, factory in self.CREATE_NEXT_VERSION_ERROR_PARAMS:
            with self.subTest(message, current=current_version_str):
                current: Version = Version.from_string(current_version_str)

                command: VersionModifier = factory(current)

                self.assertRaises(PreviewMismatchError, command.create_new_version)



#
# Copyright (c) 2021-2022 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
from .core import (
    DefaultVcsProvider,
    VcsProvider,
    VcsProviderFactory,
    VcsProviderRegistry,
    vcs_provider,
    vcs_providers,
)

# Justification: Add items to registry
from .gitcli import *  # noqa: disable=F401,F403

__all__ = (
    "vcs_provider",
    "vcs_providers",
    "VcsProvider",
    "VcsProviderFactory",
    "VcsProviderRegistry",
    "DefaultVcsProvider",
)



#
# Copyright (c) 2021-2022 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#

from typing import Final, FrozenSet

from .action import ActionCollection, VersionModifier
from .logging import logger
from .vcs import VcsProvider
from .version import Pep440VersionFormatter, Version


class _VcsSupportedVersionModifier(VersionModifier):
    def __init__(self, version: Version, vcs_provider: VcsProvider) -> None:
        super().__init__(version)
        self._vcs_provider = vcs_provider

    def create_new_version(self) -> Version:
        self._run_action()

        return self._suggest_new_version()

    def _run_action(self) -> None:
        # Empty implementation used in sub-classes
        pass

    def _suggest_new_version(self) -> Version:
        return self.current_version


class CreateTagFromVersion(_VcsSupportedVersionModifier):
    def __init__(self, version, vcs_provider, dry_run: bool) -> None:
        super().__init__(version, vcs_provider)
        self.__dry_run = dry_run

    def _run_action(self) -> None:
        if not self.__dry_run:
            self._vcs_provider.create_tag_from_version(self.current_version)
        else:
            logger.info(
                "Would create tag v%s",
                Pep440VersionFormatter().format(self.current_version),
            )


COMMAND_NAME_CREATE_TAG: Final[str] = "tag"

COMMAND_NAMES: Final[FrozenSet[str]] = frozenset(
    [
        COMMAND_NAME_CREATE_TAG,
    ]
)


def apply_vcs_based_actions(
    actions: ActionCollection, vcs_provider: VcsProvider, dry_run: bool
) -> None:
    actions[COMMAND_NAME_CREATE_TAG] = lambda v: CreateTagFromVersion(
        v, vcs_provider, dry_run
    )



#
# Copyright (c) 2021-2022 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
from typing import Optional, Protocol, Type

# MyPy cannot resolve this during pull request
from pdm.project.config import ConfigItem as _ConfigItem  # type: ignore

from .plugin import BumpCommand as _Command


class _CoreLike(Protocol):
    def register_command(
        self, command: Type[_Command], name: Optional[str] = None
    ) -> None:
        # Method empty: Only a protocol stub
        pass

    @staticmethod
    def add_config(name: str, config_item: _ConfigItem) -> None:
        # Method empty: Only a protocol stub
        pass


def main(core: _CoreLike) -> None:
    core.register_command(_Command)



#
# Copyright (c) 2021-2022 Carsten Igel.
#
# This file is part of pdm-bump 
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license. 
# Refer to LICENSE for more information
#

from unittest import TestCase
from typing import Type, Optional, Dict, List, Tuple, Iterable
from argparse import ArgumentParser, Namespace, ArgumentError

from pdm.core import Core
from pdm.project.config import ConfigItem
from pdm.cli.commands.base import BaseCommand

from pdm_bump.cli import main
from pdm_bump.plugin import BumpCommand


_registered_configurations: Dict[str, ConfigItem] = {}


class _CoreStub(object):
    registered_command: Optional[Type[BaseCommand]] = None
    registered_name: Optional[str] = None

    def register_command(self, command: Type[BaseCommand], name: Optional[str] = None) -> None:
        self.registered_command = command
        self.registered_name = name

    @staticmethod
    def add_config(name: str, config_item: ConfigItem) -> None:
        _registered_configurations[str] = config_item


class CliRegistrationTest(TestCase):
    def tearDown(self) -> None:
        _registered_configurations.clear()

    def test_registration_command_type(self) -> None:
        core: _CoreStub = _CoreStub()
        main(core)
        self.assertEqual(core.registered_command, BumpCommand, "The command is registered correctly")

    def test_registration_command_name(self) -> None:
        core: _CoreStub = _CoreStub()
        main(core)
        self.assertIsNone(core.registered_name, "No name has been supplied")

    def test_registered_config_items(self) -> None:
        core: _CoreStub = _CoreStub()
        main(core)
        self.assert_(len(_registered_configurations) == 0, "No config item has been registered")


class BumpVersionCommandParserTests(TestCase):
    SUB_TEST_PARAMS_WHAT_SUCCESS: Iterable[Tuple[str, str]] = [

        ("major", "major"),
        ("minor", "minor"),
        ("micro", "micro"),
        ("patch", "patch"),
        ("pre-release", "pre-release"),
        ("no-pre-release", "no-pre-release"),
        ("post", "post"),
        ("dev", "dev"),
    ]
    SUB_TEST_PARAMS_PRE_SUCCESS: Iterable[Tuple[str, str]] = [
        ("alpha", "alpha"),
        ("beta", "beta"),
        ("c", "c"),
        ("rc", "rc"),
    ]
    SUB_TEST_PARAMS_FAIL: Iterable[Tuple[str, List[str]]] = [
        ("major", ["maj"]),
        ("minor", ["min"]),
        ("micro", ["mic"]),
        ("patch", ["pa"]),
        ("pre-release", ["pre"]),
        ("no-pre-release", ["no-pre"]),
        ("alpha", ["pre-release", "--pre", "a"]),
        ("beta", ["pre-release", "--pre", "b"]),
        ("c", ["pre-release", "--pre", "release"]),
        ("rc", ["pre-release", "--pre", "release-candidate"]),
    ]

    def test_parser_setup_what_success(self) -> None:
        parser: ArgumentParser = ArgumentParser()
        _ = BumpCommand(parser)

        for msg, what_success in self.SUB_TEST_PARAMS_WHAT_SUCCESS:
            with self.subTest(msg, what=what_success):
                options: Namespace = parser.parse_args([ what_success ])
                self.assertEqual(options.what, what_success, "The first item is only matched")

    def test_parser_setup_pre_success(self) -> None:
        parser: ArgumentParser = ArgumentParser()
        _ = BumpCommand(parser)

        for msg, pre_success in self.SUB_TEST_PARAMS_PRE_SUCCESS:
            with self.subTest(msg, pre=pre_success):
                options: Namespace = parser.parse_args([ "pre-release", "--pre", pre_success ])
                self.assertEqual(options.pre, pre_success, "The second item is only matched")

    def test_parser_setup_fail(self) -> None:
        parser: ArgumentParser = ArgumentParser(exit_on_error=False)
        _ = BumpCommand(parser)

        for msg, options in self.SUB_TEST_PARAMS_FAIL:
            with self.subTest(msg, options=options):
                call = lambda options=options: parser.parse_args(options)
                self.assertRaises(ArgumentError, call)



#
# Copyright (c) 2021-2022 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
import sys
from abc import ABC, abstractmethod
from dataclasses import asdict as dataclass_to_dict
from typing import (
    Any,
    Callable,
    Dict,
    Final,
    FrozenSet,
    List,
    Literal,
    Tuple,
    Union,
    cast,
    final,
)

from .logging import logger, traced_function
from .version import NonNegativeInteger, Pep440VersionFormatter, Version

if sys.version_info >= (3, 10, 0):
    # suspicious mypy behavior
    from typing import TypeAlias  # type: ignore
else:
    from typing_extensions import TypeAlias

_formatter = Pep440VersionFormatter()


class PreviewMismatchError(Exception):
    pass


class VersionModifier(ABC):
    def __init__(self, version: Version) -> None:
        self.__version = version

    @property
    def current_version(self) -> Version:
        return self.__version

    @abstractmethod
    def create_new_version(self) -> Version:
        raise NotImplementedError()

    def _report_new_version(self, next_version: Version) -> None:
        logger.info(
            "Performing increment of version: %s -> %s",
            _formatter.format(self.current_version),
            _formatter.format(next_version),
        )


class _PreReleaseIncrementingVersionModified(VersionModifier):
    def __init__(self, version: Version, increment_micro: bool = True) -> None:
        super().__init__(version)
        self.__increment_micro = increment_micro

    @traced_function
    def create_new_version(self) -> Version:
        letter: Literal["a", "b", "c", "alpha", "beta", "rc"]
        name: str
        letter, name = self.pre_release_part
        pre: Tuple[
            Literal["a", "b", "c", "alpha", "beta", "rc"], NonNegativeInteger
        ] = (letter, 1)

        logger.debug(
            "Incrementing %s part of current version %s",
            name,
            _formatter.format(self.current_version),
        )

        if self.current_version.preview is not None:
            if not self._is_valid_preview_version():
                raise PreviewMismatchError(
                    f"{_formatter.format(self.current_version)} "
                    # Weird behavior of sonarlint, pylint and flake8
                    # Variable is declared as unused, if used only in
                    # formatted string
                    + "is not an "
                    + name
                    + " version."
                )
            pre = cast(
                Tuple[
                    Literal["a", "b", "c", "alpha", "beta", "rc"],
                    NonNegativeInteger,
                ],
                self.current_version.preview,
            )
            number = pre[1] + 1 if letter == pre[0] else 1

            pre = (letter, number)

        result: Version = Version(
            self.current_version.epoch,
            self._get_next_release(),
            pre,
            None,
            None,
            None,
        )

        self._report_new_version(result)

        return result

    @property
    @abstractmethod
    def pre_release_part(
        self,
    ) -> Tuple[Literal["a", "b", "c", "alpha", "beta", "rc"], str]:
        raise NotImplementedError()

    @abstractmethod
    def _is_valid_preview_version(self) -> bool:
        raise NotImplementedError()

    def _get_next_release(self) -> Tuple[NonNegativeInteger, ...]:
        micro = self.current_version.micro
        if self.__increment_micro and self.current_version.preview is None:
            micro = micro + 1

        ret: List[NonNegativeInteger] = []
        for val in self.current_version.release:
            ret.append(val)

        while len(ret) < 3:
            ret.append(0)

        ret[2] = micro

        return tuple(ret)


@final
class AlphaIncrementingVersionModifier(_PreReleaseIncrementingVersionModified):
    @property
    def pre_release_part(
        self,
    ) -> Tuple[Literal["a", "b", "c", "alpha", "beta", "rc"], str]:
        return ("a", "alpha")

    def _is_valid_preview_version(self) -> bool:
        return self.current_version.is_alpha


@final
class BetaIncrementingVersionModifier(_PreReleaseIncrementingVersionModified):
    @property
    def pre_release_part(
        self,
    ) -> Tuple[Literal["a", "b", "c", "alpha", "beta", "rc"], str]:
        return ("b", "alpha or beta")

    def _is_valid_preview_version(self) -> bool:
        return self.current_version.is_alpha or self.current_version.is_beta


@final
class ReleaseCandidateIncrementingVersionModifier(
    _PreReleaseIncrementingVersionModified
):
    @property
    def pre_release_part(
        self,
    ) -> Tuple[Literal["a", "b", "c", "alpha", "beta", "rc"], str]:
        return ("rc", "pre-release")

    def _is_valid_preview_version(self) -> bool:
        return (
            self.current_version.is_alpha
            or self.current_version.is_beta
            or self.current_version.is_release_candidate
        )


class _NonFinalPartsRemovingVersionModifier(VersionModifier):
    def __init__(self, version: Version, remove_parts: bool = True) -> None:
        super().__init__(version)
        self.__remove_parts = remove_parts

    @property
    def remove_non_final_parts(self) -> bool:
        return self.__remove_parts

    @abstractmethod
    def create_new_version(self) -> Version:
        raise NotImplementedError()

    @staticmethod
    def _create_new_constructional_args(
        release: Tuple[NonNegativeInteger, ...], epoch: NonNegativeInteger = 0
    ) -> Dict[str, Any]:
        return {
            "epoch": epoch,
            "release_tuple": release,
            "preview": None,
            "post": None,
            "dev": None,
            "local": None,
        }


_NFPR: TypeAlias = _NonFinalPartsRemovingVersionModifier


class _ReleaseVersionModifier(_NonFinalPartsRemovingVersionModifier):
    @property
    @abstractmethod
    def release_part(self) -> NonNegativeInteger:
        raise NotImplementedError()

    @traced_function
    def create_new_version(self):
        construction_args: Dict[str, Any] = dataclass_to_dict(
            self.current_version
        )

        next_release: Tuple[
            NonNegativeInteger, ...
        ] = self._update_release_version_part(self.release_part)

        if self.remove_non_final_parts:
            logger.debug("Removing non-final parts of version")
            # Using type alias due to line length enforced by black
            construction_args = _NFPR._create_new_constructional_args(  # noqa: E501 pylint: disable=W0212
                next_release, self.current_version.epoch
            )
        else:
            construction_args["release_tuple"] = next_release

        next_version: Version = Version(**construction_args)
        self._report_new_version(next_version)

        return next_version

    def _update_release_version_part(
        self, part_id: NonNegativeInteger
    ) -> Tuple[NonNegativeInteger, ...]:
        release_part: List[NonNegativeInteger] = list(
            self.current_version.release
        )

        for i in range(len(release_part)):  # pylint: disable=C0200
            logger.debug("Checking if version part at %d must be modified", i)
            if i == part_id:
                logger.debug("Incrementing version part at position %d.", i)
                release_part[i] = release_part[i] + 1
            elif i > part_id:
                logger.debug("Resetting version part at position %d", i)
                release_part[i] = 0

        return tuple(release_part)


@final
class FinalizingVersionModifier(_NonFinalPartsRemovingVersionModifier):
    def __init__(self, version: Version) -> None:
        super().__init__(version, True)

    @traced_function
    def create_new_version(self) -> Version:
        constructional_args: Dict[
            str,
            Any
            # Using type alias due to line length enforced by black
        ] = _NFPR._create_new_constructional_args(  # noqa: E501 pylint: disable=W0212
            self.current_version.release, self.current_version.epoch
        )

        next_version: Version = Version(**constructional_args)
        self._report_new_version(next_version)

        return next_version


@final
class MajorIncrementingVersionModifier(_ReleaseVersionModifier):
    __MAJOR_PART: Final[NonNegativeInteger] = 0

    @property
    def release_part(self) -> NonNegativeInteger:
        return self.__MAJOR_PART


@final
class MinorIncrementingVersionModifier(_ReleaseVersionModifier):
    __MINOR_PART: Final[NonNegativeInteger] = 1

    @property
    def release_part(self) -> NonNegativeInteger:
        return self.__MINOR_PART


@final
class MicroIncrementingVersionModifier(_ReleaseVersionModifier):
    __MICRO_PART: Final[NonNegativeInteger] = 2

    @property
    def release_part(self) -> int:
        return self.__MICRO_PART


@final
class EpochIncrementingVersionModifier(_NonFinalPartsRemovingVersionModifier):
    def __init__(
        self,
        version: Version,
        remove_parts: bool = True,
        reset_version: bool = True,
    ):
        super().__init__(version, remove_parts)
        self.__reset_version = reset_version

    @traced_function
    def create_new_version(self) -> Version:
        constructional_args: Dict[str, Any] = dataclass_to_dict(
            self.current_version
        )

        if self.__reset_version or self.remove_non_final_parts:
            constructional_args = dataclass_to_dict(Version.default())
            if not self.__reset_version:
                logger.debug("Current version tuple shall not be reset")
                constructional_args[
                    "release_tuple"
                ] = self.current_version.release

        logger.debug("Incrementing Epoch of version")
        constructional_args["epoch"] = self.current_version.epoch + 1

        next_version: Version = Version(**constructional_args)
        self._report_new_version(next_version)
        return next_version


@final
class DevelopmentVersionIncrementingVersionModifier(VersionModifier):
    @traced_function
    def create_new_version(self) -> Version:
        dev_version: NonNegativeInteger = 1
        if self.current_version.dev is not None:
            _, dev_version = self.current_version.dev
            logger.debug("Incrementing development version part by one")
            dev_version = dev_version + 1

        constructional_args: Dict[str, Any] = dataclass_to_dict(
            self.current_version
        )
        constructional_args["dev"] = ("dev", dev_version)

        next_version: Version = Version(**constructional_args)
        self._report_new_version(next_version)

        return next_version


@final
class PostVersionIncrementingVersionModifier(VersionModifier):
    @traced_function
    def create_new_version(self) -> Version:
        post_version: NonNegativeInteger = 1
        if self.current_version.post is not None:
            _, post_version = self.current_version.post
            logger.debug("Incrementing post version part by one")
            post_version = post_version + 1

        constructional_args: Dict[str, Any] = dataclass_to_dict(
            self.current_version
        )
        constructional_args["post"] = ("post", post_version)

        next_version: Version = Version(**constructional_args)
        self._report_new_version(next_version)

        return next_version


VersionModifierFactory: TypeAlias = Callable[[Version], VersionModifier]
_NestedMappingTable: TypeAlias = Dict[str, VersionModifierFactory]
ActionChoice: TypeAlias = Union[_NestedMappingTable, VersionModifierFactory]


@final
class ActionCollection(Dict[str, ActionChoice]):
    @traced_function
    def get_action(self, action: str) -> VersionModifierFactory:
        if action in self.keys():
            choice: ActionChoice = self[action]
            if not isinstance(choice, str):
                return cast(VersionModifierFactory, choice)

        raise ValueError(f"{action} is not a valid command")

    @traced_function
    def get_action_with_option(
        self, action: str, option: str
    ) -> VersionModifierFactory:
        if action in self.keys():
            choice: ActionChoice = self[action]
            if not isinstance(choice, str):
                table: _NestedMappingTable = cast(_NestedMappingTable, choice)
                if option in table.keys():
                    choice = table[option]
                    return cast(VersionModifierFactory, choice)

                raise ValueError(f"{option} is not a valid option")

        raise ValueError(f"{action} is not a valid command")


COMMAND_NAME_MAJOR_INCREMENT: Final[str] = "major"
COMMAND_NAME_MINOR_INCREMENT: Final[str] = "minor"
COMMAND_NAME_MICRO_INCREMENT: Final[str] = "micro"
COMMAND_NAME_PATCH_INCREMENT: Final[str] = "patch"
COMMAND_NAME_EPOCH_INCREMENT: Final[str] = "epoch"
COMMAND_NAME_NO_PRERELEASE: Final[str] = "no-pre-release"
COMMAND_NAME_PRERELEASE: Final[str] = "pre-release"
COMMAND_NAME_DEV_INCREMENT: Final[str] = "dev"
COMMAND_NAME_POST_INCREMENT: Final[str] = "post"

PRE_RELEASE_OPTION_ALPHA: Final[str] = "alpha"
PRE_RELEASE_OPTION_BETA: Final[str] = "beta"
PRE_RELEASE_OPTION_RC: Final[str] = "rc"
PRE_RELEASE_OPTION_RC_ALT: Final[str] = "c"

_next_major: TypeAlias = MajorIncrementingVersionModifier
_next_minor: TypeAlias = MinorIncrementingVersionModifier
_next_micro: TypeAlias = MicroIncrementingVersionModifier
_next_alpha: TypeAlias = AlphaIncrementingVersionModifier
_next_beta: TypeAlias = BetaIncrementingVersionModifier
_next_rc: TypeAlias = ReleaseCandidateIncrementingVersionModifier
_no_pre: TypeAlias = FinalizingVersionModifier
_next_dev: TypeAlias = DevelopmentVersionIncrementingVersionModifier
_next_post: TypeAlias = PostVersionIncrementingVersionModifier
_next_epoch: TypeAlias = EpochIncrementingVersionModifier


def create_actions(
    *,
    remove_parts: bool = True,
    reset_version: bool = True,
    increment_micro: bool = False,
) -> ActionCollection:
    return ActionCollection(
        {
            COMMAND_NAME_MAJOR_INCREMENT: lambda v: _next_major(
                v, remove_parts
            ),
            COMMAND_NAME_MINOR_INCREMENT: lambda v: _next_minor(
                v, remove_parts
            ),
            COMMAND_NAME_MICRO_INCREMENT: lambda v: _next_micro(
                v, remove_parts
            ),
            COMMAND_NAME_PATCH_INCREMENT: lambda v: _next_micro(
                v, remove_parts
            ),
            COMMAND_NAME_PRERELEASE: {
                PRE_RELEASE_OPTION_ALPHA: lambda v: _next_alpha(
                    v, increment_micro
                ),
                PRE_RELEASE_OPTION_BETA: lambda v: _next_beta(
                    v, increment_micro
                ),
                PRE_RELEASE_OPTION_RC: lambda v: _next_rc(v, increment_micro),
                PRE_RELEASE_OPTION_RC_ALT: lambda v: _next_rc(
                    v, increment_micro
                ),
            },
            # Justifications: Explicit call of type, line too long due to noqa
            COMMAND_NAME_NO_PRERELEASE: lambda v: _no_pre(  # noqa: E501 pylint: disable=W0108
                v
            ),
            COMMAND_NAME_EPOCH_INCREMENT: lambda v: _next_epoch(
                v, remove_parts, reset_version
            ),
            # Justifications: Explicit call of type, line too long due to noqa
            COMMAND_NAME_DEV_INCREMENT: lambda v: _next_dev(  # noqa: E501 pylint: disable=W0108
                v
            ),
            # Justifications: Explicit call of type, line too long due to noqa
            COMMAND_NAME_POST_INCREMENT: lambda v: _next_post(  # noqa: E501 pylint: disable=W0108
                v
            ),
        }
    )


COMMAND_NAMES: Final[FrozenSet[str]] = frozenset(
    [
        COMMAND_NAME_MAJOR_INCREMENT,
        COMMAND_NAME_MINOR_INCREMENT,
        COMMAND_NAME_MICRO_INCREMENT,
        COMMAND_NAME_PATCH_INCREMENT,
        COMMAND_NAME_PRERELEASE,
        COMMAND_NAME_NO_PRERELEASE,
        COMMAND_NAME_EPOCH_INCREMENT,
        COMMAND_NAME_DEV_INCREMENT,
        COMMAND_NAME_POST_INCREMENT,
    ]
)

PRERELEASE_OPTIONS: Final[FrozenSet[str]] = frozenset(
    [
        PRE_RELEASE_OPTION_ALPHA,
        PRE_RELEASE_OPTION_BETA,
        PRE_RELEASE_OPTION_RC,
        PRE_RELEASE_OPTION_RC_ALT,
    ]
)



#
# Copyright (c) 2021-2022 Carsten Igel, Chase Sterling.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, cast

from .config import Config
from .version import Pep440VersionFormatter, Version

DEFAULT_REGEX = re.compile(
    r"^__version__\s*=\s*[\"'](?P<version>.+?)[\"']\s*(?:#.*)?$", re.M
)


@dataclass
class DynamicVersionConfig:
    file: Path
    regex: re.Pattern


class DynamicVersionSource:
    def __init__(self, project_root: Path, config: Config) -> None:
        self.__project_root = project_root
        self.__config = config
        self.__current_version: Optional[Version] = None

    @property
    def is_enabled(self) -> bool:
        dynamic_items: Optional[list[str]] = cast(
            list[str], self.__config.get_pyproject_value("project", "dynamic")
        )
        return dynamic_items is not None and "version" in dynamic_items

    def __get_current_version(self) -> Version:
        if self.__current_version is not None:
            return cast(Version, self.__current_version)

        dynamic: DynamicVersionConfig = self.__get_dynamic_version()
        version: Optional[str] = __get_dynamic_version(dynamic)
        if version is not None:
            self.__current_version = Version.from_string(version)
            return self.__current_version
        raise ValueError(
            f"Failed to find version in {dynamic.file}. "
            f"Make sure it matches {dynamic.regex}"
        )

    def __set_current_version(self, version: Version) -> None:
        self.__current_version = version

    current_version = property(__get_current_version, __set_current_version)

    def save_value(self) -> None:
        if self.__current_version is None:
            raise ValueError("No current value set")
        version: Version = cast(Version, self.__current_version)
        ver: str = Pep440VersionFormatter().format(version)
        config: DynamicVersionConfig = self.__get_dynamic_version()
        __replace_dynamic_version(config, ver)

    def __get_dynamic_version(self) -> DynamicVersionConfig:
        dynamic_version: Optional[
            DynamicVersionConfig
        ] = __find_dynamic_config(self.__project_root, self.__config)
        if dynamic_version is not None:
            dynamic: DynamicVersionConfig = cast(
                DynamicVersionConfig, dynamic_version
            )
            return dynamic
        raise ValueError(
            f"Failed to extract dynamic version from {self.__project_root}."
            f" Only pdm-pep517 `file` types are supported."
        )


def __find_dynamic_config(
    root_path: Path, project_config: Config
) -> Optional[DynamicVersionConfig]:
    if (
        project_config.get_pyproject_value("build-system", "build-backend")
        == "pdm.pep517.api"
        and project_config.get_pyproject_value(
            "tool", "pdm", "version", "source"
        )
        == "file"
    ):
        file_path = project_config.get_pyproject_value(
            "tool", "pdm", "version", "path"
        )
        return DynamicVersionConfig(
            file=root_path / file_path,
            regex=DEFAULT_REGEX,
        )
    return None


def __get_dynamic_version(config: DynamicVersionConfig) -> Optional[str]:
    with config.file.open("r") as file_ptr:
        match = config.regex.search(file_ptr.read())
    if match is not None:
        return match.group("version")
    return None


def __replace_dynamic_version(
    config: DynamicVersionConfig, new_version: str
) -> None:
    with config.file.open("r") as file_ptr:
        version_file = file_ptr.read()
        match = config.regex.search(version_file)
        if match is None:
            raise ValueError("Failed to fetch version")
        match = cast(re.Match[str], match)
        version_start, version_end = match.span("version")
        new_version_file = (
            version_file[:version_start]
            + new_version
            + version_file[version_end:]
        )
    with config.file.open("w") as file_ptr:
        file_ptr.write(new_version_file)



#
# Copyright (c) 2021-2022 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
from argparse import ArgumentParser, Namespace
from pathlib import Path
from traceback import format_exc as get_traceback
from typing import Final, Optional, Protocol, cast, final

# MyPy does not recognize this during pull requests
from pdm.cli.commands.base import BaseCommand  # type: ignore
from pdm.termui import UI  # type: ignore

from .action import COMMAND_NAMES as MODIFIER_ACTIONS
from .action import (
    PRERELEASE_OPTIONS,
    ActionCollection,
    VersionModifier,
    VersionModifierFactory,
    create_actions,
)
from .auto import COMMAND_NAMES as VCS_BASED_ACTIONS
from .auto import apply_vcs_based_actions
from .config import Config, ConfigHolder
from .dynamic import DynamicVersionSource
from .logging import logger, traced_function, update_logger_from_project_ui
from .source import StaticPep621VersionSource
from .vcs import (
    DefaultVcsProvider,
    VcsProvider,
    VcsProviderRegistry,
    vcs_providers,
)
from .version import Pep440VersionFormatter, Version


# Justification: Protocol for interoperability
class _CoreLike(Protocol):  # pylint: disable=R0903
    ui: UI


class _ProjectLike(ConfigHolder, Protocol):
    root: Path
    core: _CoreLike

    @property
    def pyproject_file(self) -> str:
        # Method empty: Only a protocol stub
        pass

    def write_pyproject(self, show_message: bool) -> None:
        # Method empty: Only a protocol stub
        pass


class _VersionSource(Protocol):
    @property
    def is_enabled(self) -> bool:
        raise NotImplementedError()

    def __get_current_version(self) -> Version:
        raise NotImplementedError()

    def __set_current_version(self, version: Version) -> None:
        raise NotImplementedError()

    current_version: property = property(
        __get_current_version, __set_current_version
    )

    def save_value(self) -> None:
        raise NotImplementedError()


@final
class BumpCommand(BaseCommand):
    name: Final[str] = "bump"
    description: str = "Bumps the version to a next version following PEP440."

    @traced_function
    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "what",
            action="store",
            choices=list(MODIFIER_ACTIONS) + list(VCS_BASED_ACTIONS),
            default=None,
            help="The part of the version to bump according to PEP 440: "
            + "major.minor.micro.",
        )
        parser.add_argument(
            "--pre",
            action="store",
            choices=list(PRERELEASE_OPTIONS),
            default=None,
            help="Sets a pre-release on the current version."
            + " If a pre-release is set, it can be removed "
            + "using the final option. A new pre-release "
            + "must greater then the current version."
            + " See PEP440 for details.",
        )
        parser.add_argument(
            "--dry-run",
            "-n",
            action="store_true",
            help="Do not store incremented version",
        )
        parser.add_argument(
            "--micro",
            action="store_true",
            help="When setting pre-release, specifies "
            + "whether micro version shall "
            + "be incremented as well",
        )
        parser.add_argument(
            "--reset",
            action="store_true",
            help="When setting epoch, reset version to 1.0.0",
        )
        parser.add_argument(
            "--no-remove",
            action="store_false",
            help="When incrementing major, minor, micro or epoch, "
            + "do not remove all pre-release parts",
        )

    @traced_function
    def handle(self, project: _ProjectLike, options: Namespace) -> None:
        config: Config = Config(project)
        update_logger_from_project_ui(project.core.ui)

        selected_backend: Optional[_VersionSource] = self._select_backend(
            project, config
        )

        if selected_backend is None:
            logger.error("Cannot find version in %s", project.pyproject_file)
            return

        backend: _VersionSource = cast(_VersionSource, selected_backend)

        version: Version = backend.current_version

        next_version: Version = self._get_next_version(
            version, project, options
        )

        if next_version == version:
            logger.info(
                "Version did not change after application. "
                "No need to persist new version."
            )
            return

        if options.dry_run:
            logger.info("Would write new version %s", next_version)
            return

        self._save_new_version(backend, next_version)

    @traced_function
    def _get_vcs_provider(self, project: _ProjectLike) -> VcsProvider:
        config: Config = Config(project)
        value = config.get_config_or_pyproject_value(
            "pdm_bump", "vcs", "provider"
        )

        registry: VcsProviderRegistry = vcs_providers

        if value is not None:
            provider_name: str = str(value)
            factory_method = registry[provider_name]
            if factory_method is not None:
                factory = factory_method()
                return factory.force_create_provider(project.root)
        else:
            provider = registry.find_repository_root(project.root)
            if provider is not None:
                return provider

        return DefaultVcsProvider(project.root)

    @traced_function
    def _get_next_version(
        self, version: Version, project: _ProjectLike, options: Namespace
    ) -> Version:
        actions: ActionCollection = self._get_actions(
            options.micro, options.reset, options.no_remove
        )

        vcs_provider: VcsProvider = self._get_vcs_provider(project)
        apply_vcs_based_actions(actions, vcs_provider, options.dry_run)

        modifier: VersionModifier = self._get_action(
            actions, version, options.what, options.pre
        )

        try:
            return modifier.create_new_version()
        except ValueError as exc:
            logger.exception(
                "Failed to update version to next version", exc_info=False
            )
            logger.debug("Exception occurred: %s", get_traceback())
            raise SystemExit(1) from exc

    @traced_function
    def _select_backend(
        self, project: _ProjectLike, config: Config
    ) -> Optional[_VersionSource]:
        static_backend: _VersionSource = StaticPep621VersionSource(
            project, config
        )
        dynamic_backend: _VersionSource = DynamicVersionSource(
            project.root, config
        )

        selected_backend: Optional[_VersionSource] = None
        for backend in (static_backend, dynamic_backend):
            if backend.is_enabled:
                selected_backend = backend
                break

        return selected_backend

    @traced_function
    def _save_new_version(
        self, backend: _VersionSource, next_version: Version
    ) -> None:
        backend.current_version = next_version

        backend.save_value()

    @traced_function
    def _get_action(
        self,
        actions: ActionCollection,
        version: Version,
        what: str,
        pre: Optional[str],
    ) -> VersionModifier:
        modifier_factory: VersionModifierFactory
        if pre is not None:
            modifier_factory = actions.get_action_with_option(what, pre)
        else:
            modifier_factory = actions.get_action(what)

        modifier: VersionModifier = modifier_factory(version)
        return modifier

    @traced_function
    def _get_actions(
        self, increment_micro: bool, reset_version: bool, remove_parts: bool
    ) -> ActionCollection:
        actions: ActionCollection = create_actions(
            increment_micro=increment_micro,
            reset_version=reset_version,
            remove_parts=remove_parts,
        )

        return actions

    @traced_function
    def _version_to_string(self, version: Version) -> str:
        result: str = Pep440VersionFormatter().format(version)
        return result



#
# Copyright (c) 2021-2022 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#

from typing import Protocol, cast

from .config import Config
from .version import Pep440VersionFormatter, Version


# Justification: Minimal protocol
class _ProjectWriter(Protocol):  # pylint: disable=R0903
    def write_pyproject(self, show_message: bool) -> None:
        # Method empty: Only a protocol stub
        pass


class StaticPep621VersionSource:
    def __init__(self, project: _ProjectWriter, config: Config) -> None:
        self.__project = project
        self.__config = config

    @property
    def is_enabled(self) -> bool:
        return (
            self.__config.get_pyproject_value("project", "version") is not None
        )

    def __get_current_version(self) -> Version:
        version: str = cast(
            str, self.__config.get_pyproject_value("project", "version")
        )
        return Version.from_string(version)

    def __set_current_version(self, ver: Version) -> None:
        formatter: Pep440VersionFormatter = Pep440VersionFormatter()
        version: str = formatter.format(ver)
        self.__config.set_pyproject_value(version, "project", "version")

    current_version = property(__get_current_version, __set_current_version)

    def save_value(self) -> None:
        self.__project.write_pyproject(True)



#
# Copyright (c) 2021-2022 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
from typing import Any, Optional, Protocol, cast

from .logging import logger, traced_function


class ConfigHolder(Protocol):
    @property
    def pyproject(self) -> dict[str, Any]:
        # Method empty: Only a protocol stub
        pass

    @property
    def config(self) -> dict[str, Any]:
        # Method empty: Only a protocol stub
        pass


@traced_function
def _get_config_value(config: dict[str, Any], *keys: str) -> Optional[Any]:
    front: str

    key: str = ".".join(keys)
    logger.debug("Searching for '%s' in configuration", key)
    logger.debug("Configuration is set to: \n%s", config)

    while len(keys) > 1:
        front = keys[0]
        if front in config.keys():
            logger.debug("Found configuration section %s", front)
            config = cast(dict[str, Any], config[front])
            keys = tuple(keys[1:])
        else:
            logger.debug("Could not find configuration section %s.", front)
            return None

    front = keys[0]

    result = None if front not in config.keys() else config[front]
    logger.debug("Found value at '%s' is: %s", key, result)
    return result


@traced_function
def _set_config_value(config: dict[str, Any], value: Any, *keys: str) -> None:
    front: str

    key: str = ".".join(keys)
    logger.debug("Setting '%s' to '%s'", key, value)

    while len(keys) > 1:
        front = keys[0]
        if front not in config.keys():
            logger.debug(
                "Key '%s' was not found. Adding empty configuration", front
            )
            config[front] = {}
        config = config[front]
        keys = tuple(keys[1:])

    front = keys[0]
    config[front] = value


class Config:
    def __init__(self, project: ConfigHolder) -> None:
        self.__project: ConfigHolder = project

    @traced_function
    def get_pyproject_value(self, *keys: str) -> Optional[Any]:
        config: dict[str, Any] = self.__project.pyproject
        return _get_config_value(config, *keys)

    @traced_function
    def get_config_value(self, *keys: str) -> Optional[Any]:
        config: dict[str, Any] = self.__project.config
        return _get_config_value(config, *keys)

    @traced_function
    def get_config_or_pyproject_value(self, *keys: str) -> Optional[Any]:
        config1: dict[str, Any] = self.__project.config
        config2: dict[str, Any] = self.__project.pyproject
        py_project_keys: list[str] = ["tool", "pdm", "plugins"] + list(keys)

        return _get_config_value(config1, *keys) or _get_config_value(
            config2, *tuple(py_project_keys)
        )

    @traced_function
    def set_pyproject_value(self, value: Any, *keys: str) -> None:
        config: dict[str, Any] = self.__project.pyproject
        _set_config_value(config, value, *keys)
